
/***************************************************************************
 *   This file is part of Aspect, a simple PEC tool.                       *
 *                                                                         *
 *   Copyright (C) 2007 by Wolfgang Hoffmann <woho@woho.de>                *
 *                                                                         *
 *   This program is free software, licensed under the GPL v2.             *
 *   See the file COPYING for more details.                                *
 ***************************************************************************/

#include "data.h"
#include "log.h"

#include <QtCore>
#include <math.h>




/********************************************************************/
/*!
    \class PecSample
    \brief Data container for one PEC table entry
    \sa PecData
*/

/*!
    \enum PecSample::Val
    Possible values for one PEC table entry
*/


int PecSample::val2Int(PecSample::Val eVal)
    {
    switch (eVal)
        {
        default:                return 0;
        case PecSample::faster: return 1;
        case PecSample::slower: return 2;
        case PecSample::reset:  return 3;
        }
    }


PecSample::Val PecSample::int2Val(int nVal)
    {
    switch (nVal)
        {
        default: return PecSample::keep;
        case 1:  return PecSample::faster;
        case 2:  return PecSample::slower;
        case 3:  return PecSample::reset;
        }
    }


int PecSample::val2Acc(PecSample::Val eVal)
    {
    switch (eVal)
        {
        default:                return  0;
        case PecSample::faster: return  1;
        case PecSample::slower: return -1;
        }
    }


PecSample::Val PecSample::acc2Val(int nAcc)
    {
    switch (nAcc)
        {
        default: return PecSample::keep;
        case  1: return PecSample::faster;
        case -1: return PecSample::slower;
        }
    }


PecSample::PecSample()
:   m_eVal(PecSample::keep),
    m_nAcc(0),
    m_nSpd(0),
    m_nPos(0)
    {
    }


PecSample::PecSample(PecSample::Val eVal)
:   m_eVal(eVal),
    m_nAcc(0),
    m_nSpd(0),
    m_nPos(0)
    {
    }


PecSample::PecSample(PecSample::Val eVal, int nAcc, int nSpd, int nPos)
:   m_eVal(eVal),
    m_nAcc(nAcc),
    m_nSpd(nSpd),
    m_nPos(nPos)
    {
    }




/********************************************************************/
/*!
    \class PecNode
    \internal
*/

class PecNode
    {
public:
    PecNode(int nRow, int nSpd, int nPos, double dfPosRef)
        : m_nRow(nRow), m_nSpd(nSpd), m_nPos(nPos),
          m_dfPosRef(dfPosRef), m_dfWeight(-1.) {};
    int row() const { return m_nRow; };
    int spd() const { return m_nSpd; };
    int pos() const { return m_nPos; };
    void linkAdd(PecNode *pNodePrev);
    void linkAddReachable(PecNode *pNodePrev);
    bool linked() const { return (m_lstPrev.size() > 0); };
    void initWeight(bool bValid);
    void calcWeight();
    double weight() const { return m_dfWeight; };
    PecNode *bestPrev() const;
    QList<PecNode *> prev() const { return m_lstPrev; };
    QString debug() const;
private:
    int m_nRow;
    int m_nSpd;
    int m_nPos;
    double m_dfPosRef;
    double m_dfWeight;
    QList<PecNode *> m_lstPrev;
    };


void PecNode::linkAdd(PecNode *pNodePrev)
    {
    if (pNodePrev == 0)
        return;
    m_lstPrev.append(pNodePrev);
    }


/*!
    Link this node with \a pNodePrev if this point can be
    reached from \a pNodePrev:
    - check if this position is reached when moving from previous position with previous speed
    - check if this speed can be reached by PEC table means, i.e. this speed is 0 (PecSample::reset)
      or within +/-1 from previous speed (PecSample::keep, PecSample::slower, PecSample::faster)
*/

void PecNode::linkAddReachable(PecNode *pNodePrev)
    {
    if (pNodePrev == 0)
        return;
    if ((m_nPos == pNodePrev->pos() + pNodePrev->spd()) &&
        ((m_nSpd == 0) || (abs(m_nSpd - pNodePrev->spd()) <= 1)))
        m_lstPrev.append(pNodePrev);
    }


void PecNode::initWeight(bool bValid)
    {
    m_dfWeight = (bValid)? 0.: 1e10;
    }


void PecNode::calcWeight()
    {
    double dfDiff = (double)m_nPos - m_dfPosRef;
    double dfWeightPrev = 1e10;
    for (int nPrev = 0; nPrev < m_lstPrev.size(); nPrev++)
        if (dfWeightPrev > m_lstPrev.at(nPrev)->weight())
            dfWeightPrev = m_lstPrev.at(nPrev)->weight();
    m_dfWeight = dfDiff * dfDiff + dfWeightPrev;
    }


PecNode *PecNode::bestPrev() const
    {
    PecNode *pNodePrev = 0;
    double dfWeightPrev = 1e10;
    for (int nPrev = 0; nPrev < m_lstPrev.size(); nPrev++)
        if (dfWeightPrev > m_lstPrev.at(nPrev)->weight())
            {
            dfWeightPrev = m_lstPrev.at(nPrev)->weight();
            pNodePrev = m_lstPrev.at(nPrev);
            }
    return pNodePrev;
    }


QString PecNode::debug() const
    {
    QString qsDebug = QString::fromUtf8("   row %1, pos %2, spd %3:")
        .arg(m_nRow, 3)
        .arg(m_nPos, 3)
        .arg(m_nSpd, 2);
    for (int nPrev = 0; nPrev < m_lstPrev.size(); nPrev++)
        qsDebug += QString::fromUtf8(" (%1%2%3)")
            .arg(m_lstPrev.at(nPrev)->pos())
            .arg((m_lstPrev.at(nPrev)->spd() >= 0)? QString::fromUtf8("+"): QString())
            .arg(m_lstPrev.at(nPrev)->spd());
    return qsDebug;
    }


class PecNodePtr
    {
public:
    PecNodePtr() : m_pNode(0) {};
    PecNodePtr(const PecNode *pNode) : m_pNode(pNode) {};
    const PecNode *node() const { return m_pNode; };
    bool operator<(const PecNodePtr &oNodePtr) const;
private:
    const PecNode *m_pNode;
    };


bool PecNodePtr::operator<(const PecNodePtr &oNodePtr) const
    {
    if (m_pNode == 0)
        return false;
    if (oNodePtr.m_pNode == 0)
        return true;
    return (m_pNode->weight() < oNodePtr.m_pNode->weight());
    }




/********************************************************************/
/*!
    \class PecData
    \brief Data container for PEC table

    This class holds data for PEC (periodic error correction).
*/


PecData::PecData()
    {
    m_aTable.resize(c_nRowNum);
    clear();
    }


/*!
    Clear PEC table

    Sets first PEC value to PecSample::reset, all other to PecSample::keep.
*/

void PecData::clear()
    {
    m_nFactor = c_nFactorMin;
    for (int nRow = 0; nRow < c_nRowNum; nRow++)
        m_aTable[nRow] = PecSample();
    m_aTable[0].m_eVal = PecSample::reset;
    m_nOffset = 0;
    }


PecData::Traversal PecData::traversal() const
    {
    return m_eTraversal;
    }


void PecData::setTraversal(Traversal eDirection)
    {
    m_eTraversal = eDirection;
    recalcGeo();
    }


int PecData::factor() const
    {
    return m_nFactor;
    }


const PecSample &PecData::operator[](int nRow) const
    {
    static PecSample oNull;
    if ((nRow < 0) || (nRow >= c_nRowNum))
        return oNull;

    return m_aTable[nRow];
    }


PecSample::Val PecData::value(int nRow) const
    {
    return operator[](nRow).m_eVal;
    }


int PecData::valueInt(int nRow) const
    {
    return PecSample::val2Int(value(nRow));
    }


bool PecData::set(int nFactor, const QVector<PecSample::Val> &aTable)
    {
    if ((nFactor < c_nFactorMin) || (nFactor > c_nFactorMax))
        return false;
    if (aTable.size() != c_nRowNum)
        return false;

    m_nFactor = nFactor;
    for (int nRow = 0; nRow < c_nRowNum; nRow++)
        m_aTable[nRow].m_eVal = aTable[nRow];

    recalcGeo();
    return true;
    }


bool PecData::setFactor(int nFactor)
    {
    if ((nFactor < c_nFactorMin) || (nFactor > c_nFactorMax))
        return false;

    m_nFactor = nFactor;
    return true;
    }


bool PecData::setValue(int nRow, PecSample::Val eVal)
    {
    if ((nRow < 0) || (nRow >= c_nRowNum))
        return false;

    m_aTable[nRow].m_eVal = eVal;
    return true;
    }


void PecData::setFinish()
    {
    recalcGeo();
    }


/*!
    Calculate PEC table to achieve the given PE profile \a aProfile

    Background:

    PE is a positional error of the worm gear.

    Let \c x denote the rotation angle of the worm, running from \c 0 ... \c 256
    for one worm revolution (360°).<br>
    Let \c y denote the rotation angle of the worm wheel, running from \c 0 ... \c 256
    when advancing by one worm wheel tooth (360° / number of worm wheel teeth).

    Ideally, <tt>y = x + C</tt>, with C being an arbitrary constant
    (worm and worm wheel beeing endless, so there's no special start point).<br>
    With PE, <tt>y = x + e(x) + C</tt>.

    To compensate the PE, the MCU models e(x) with an approximation c(x), which has
    discrete slope values. The MCU holds a PEC table of 256 entries and a PEC factor f,
    which describe <tt>c(x)</tt>. A PEC table entry has one of the following values:
    - \c 0: slope of c(x) is unchanged,<br>
        i.e. <tt>c(x+1) - c(x) = c(x) - c(x-1)</tt>
    - \c 1: slope of c(x) increases by a, if slope is \< f, otherwise is unchanged,<br>
        i.e. <tt>c(x+1) - c(x) = c(x) - c(x-1) + a</tt>
    - \c 2: slope of c(x) decreases by a, if slope is \> -f, otherwise is unchanged,<br>
        i.e. <tt>c(x+1) - c(x) = c(x) - c(x-1) - a</tt>
    - \c 3: slope of c(x) is set to 0,<br>
        i.e. <tt>c(x+1) - c(x) = 0</tt>

    c(x) is dependent on the traversal direction, i.e. whether the
    PEC table is processed with increasing or decreasing x.
    When traversing backwards, the MCU uses the PEC table entry at x to
    calculate the speed while moving from c(x+1) to c(x), i.e. it does a kind
    of look-ahead.
    It seems that currently the MCU always traverses the PEC table with
    decreasing x, so switching the RA direction (from Normal to Reversed or vice versa)
    or switching the Hemisphere invalidates the PEC table.

    Concerning \c a, empirical measurements suggest a value of 0.0125

    \note Above description of PEC encoding is an empirical guess, which
        works for my settings. Some details or even the whole principle
        might be wrong for general usage ...

    \todo Currently the algorithm only considers a small range of values (the current,
        previous and next). That seems ok for smooth PE curves, but probably fails for
        PE with sharp jumps. So find an algorithm that looks at the whole picture and
        minimizes the mean square error.

    \sa recalcGeo()
*/

bool PecData::calc(const QVector<double> &aProfile, IProgress *pProgress)
    {
    LOG(4, "PecData::calc(): traversal %d", (int)m_eTraversal);
    if (aProfile.size() != c_nRowNum)
        {
        LOG(0, "PecData::calc(): aProfile has wrong size %d instead of %d",
            aProfile.size(), c_nRowNum);
        return false;
        }

#if 0
    // traverse three times through the table:
    // - first time to get a correct start speed
    // - second time to fill in acceleration table
    // - third time as a consistency check if values are stable
    //   and to find PEC factor and minimum and maximum of resulting positions
    // this seems to work better than starting from a dedicated point
    // such as the profile minimum or maximum, which in theory has 0
    // profile slope
    int nSpd = 0;
    int nPos = (int)rint(aProfile[0]);
    int nPosMin = 1000;
    int nPosMax = -1000;
    m_nFactor = 1;
    for (int nRound = 0; nRound < 3; nRound++)
        {
        for (int nCnt = 0; nCnt < c_nRowNum; nCnt++)
            {
            // careful:
            // The firmware seems to model a PEC table entry not as a point,
            // but as an interval. The interval covers the area of the worm
            // where pecByte*4+pecBit is constant. Physical properties are
            // modeled like this:
            // - acceleration is a spike at the interval entry point, i.e.
            //   it's real position on the worm depends on traversal direction
            // - speed changes immediately at interval entry, since accel is a spike
            //   so speed is the same over the interval, independently of traversal
            //   direction, except if previous speed differs
            // - position reaches resulting value when exiting the interval,
            //   after moving from current to next point with current speed
            // So when looking at a PEC table entry point, it's values depend
            // on traversal direction:
            // - forward (this=n, prev=n-1, next=n+1):
            //   speed[this] = speed[prev] + acc[this]
            //   pos[this] = pos[prev] + speed[prev]
            // - backward (this=n, prev=n+1, next=n-1):
            //   speed[this] = speed[prev] + acc[next]
            //   pos[this] = pos[prev] + speed[prev]

            // nSpd and nPos are speed and position at entry of current interval
            int nRowThis = (m_eTraversal * nCnt + c_nRowNum) % c_nRowNum;
            int nRowNext = (m_eTraversal * (nCnt + 1) + c_nRowNum * 2) % c_nRowNum;
            int nRowAcc = (m_eTraversal == forward)? nRowThis: nRowNext;
            // dfMidProfile is position profile at half way between this and next point.
            // dfMidDiff is dfProfile minus reached position if continuing with current speed
            double dfMidProfile = (aProfile[nRowNext] + aProfile[nRowThis]) / 2.;
            double dfMidDiff = dfMidProfile - ((double)nPos + (double)nSpd / 2.);
            // calculate needed acceleration to minimize positional error
            int nAcc = 0;
            if (dfMidDiff > .5)
                nAcc = 1;
            else if (dfMidDiff < -.5)
                nAcc = -1;
            // adjust speed by acceleration and keep within limits
            nSpd += nAcc;
            if (nSpd < - c_nFactorMax)
                nSpd = - c_nFactorMax;
            if (nSpd > c_nFactorMax)
                nSpd = c_nFactorMax;
            // adjust position by speed
            nPos += nSpd;
            LOG(4, "round %d, row %d: profile %.3f, pos %d, spd %d",
                nRound, nRowThis, dfMidProfile, nPos, nSpd);
            // depending on round, write acceleration into table
            // or perform consistency check
            if (nRound == 1)
                m_aTable[nRowAcc] = PecSample::acc2Val(nAcc);
            else if (nRound == 2)
                {
                if (m_nFactor < abs(nSpd))
                    m_nFactor = abs(nSpd);
                if (nPosMin > nPos)
                    nPosMin = nPos;
                if (nPosMax < nPos)
                    nPosMax = nPos;
                if (nAcc != PecSample::val2Acc(m_aTable[nRowAcc].m_eVal))
                    LOG(0, "PecData::calc(): inconsistent result at row %d: %d != %d",
                        nRowAcc, nAcc, PecSample::val2Acc(m_aTable[nRowAcc].m_eVal));
                // set table to PecSample::reset if acc and spd are 0
                if ((nAcc == 0) && (nSpd == 0))
                    m_aTable[nRowAcc].m_eVal = PecSample::reset;
                }
            }
        }
    m_nOffset = (nPosMax + nPosMin) / 2;
#else
    // this might take a while, so do proper progress indication
    double dfProgressMax = (pProgress != 0)? (double)pProgress->maxValue(): 100.;
    double dfProgressGraphBeg = 0.;
    double dfProgressGraphEnd = 0.1;
    double dfProgressNodesBeg = 0.1;
    double dfProgressNodesEnd = 0.3;
    double dfProgressFloodBeg = 0.3;
    double dfProgressFloodEnd = 0.99;
    double dfProgressPecBeg = 0.99;
    double dfProgressPecEnd = 1.;
    if (pProgress != 0)
        pProgress->setValue(0);

    // Find PEC table with minimal square error between PE and PEC curve.
    // To do so, use a semi-brute-force approach ...

    // A node is a PEC (point,position,speed) triplet.
    // Make directed graph of these nodes, with PEC point as level criterion,
    // and breaking up the PEC cycle at PEC point 0 by duplicating PEC point 0
    // as graph top and bottom. Graph direction is PEC traversal direction,
    // i.e. level increments as PEC table is traversed
    // Hold graph in array of PEC point node lists.
    QList<PecNode> aPecPoint[c_nRowNum + 1];

    // Populate graph with nodes representing all positions within
    // profile range + c_nFactorMax and all speeds between +/- c_nFactorMax
    LOG(4, "populating graph ...");
    for (int nLevel = 0; nLevel < c_nRowNum + 1; nLevel++)
        {
        if (pProgress != 0)
            if (pProgress->setValue((int)(dfProgressMax * (dfProgressGraphBeg +
                (dfProgressGraphEnd - dfProgressGraphBeg) * (double)nLevel / (double)c_nRowNum))))
                return false;
        int nRowThis = (m_eTraversal * nLevel + c_nRowNum) % c_nRowNum;
        int nRowPrev = (m_eTraversal * (nLevel - 1) + 2 * c_nRowNum) % c_nRowNum;
        int nRowNext = (m_eTraversal * (nLevel + 1) + 2 * c_nRowNum) % c_nRowNum;
        double dfProfileMin = aProfile[nRowThis];
        double dfProfileMax = dfProfileMin;
        if (dfProfileMin > aProfile[nRowPrev])
            dfProfileMin = aProfile[nRowPrev];
        if (dfProfileMin > aProfile[nRowNext])
            dfProfileMin = aProfile[nRowNext];
        if (dfProfileMax < aProfile[nRowPrev])
            dfProfileMax = aProfile[nRowPrev];
        if (dfProfileMax < aProfile[nRowNext])
            dfProfileMax = aProfile[nRowNext];
        int nProfileMin = (int)floor(dfProfileMin) - c_nFactorMax;
        int nProfileMax = (int)ceil(dfProfileMax) + c_nFactorMax;
        for (int nPos = nProfileMin; nPos <= nProfileMax; nPos++)
            for (int nSpd = -c_nFactorMax; nSpd <= c_nFactorMax; nSpd++)
                aPecPoint[nLevel].append(PecNode(nRowThis, nSpd, nPos, aProfile[nRowThis]));
        }

    // connect each node to all nodes of prev level that are within reach of
    // one PEC step (given current and prev position and speed as well as the
    // 4 possible PEC table accelerations)
    LOG(4, "connecting nodes ...");
    for (int nLevelThis = 1; nLevelThis < c_nRowNum + 1; nLevelThis++)
        {
        if (pProgress != 0)
            if (pProgress->setValue((int)(dfProgressMax * (dfProgressNodesBeg +
                (dfProgressNodesEnd - dfProgressNodesBeg) * (double)nLevelThis / (double)c_nRowNum))))
                return false;
        int nLevelPrev = nLevelThis - 1;
        for (int nNodePrev = 0; nNodePrev < aPecPoint[nLevelPrev].size(); nNodePrev++)
            {
            // shortcut if prev node is not linked to any previous
            if ((nLevelPrev > 0) && (!aPecPoint[nLevelPrev].at(nNodePrev).linked()))
                continue;
            for (int nNodeThis = 0; nNodeThis < aPecPoint[nLevelThis].size(); nNodeThis++)
                aPecPoint[nLevelThis][nNodeThis].linkAddReachable(&(aPecPoint[nLevelPrev][nNodePrev]));
            }
        }
    // Now that all possible connections are known, find best route,
    // i.e. route that minimizes square error between PE profile and PEC curve.
    // To do so, for each bottom node flood graph to calculate error weights
    // of best route from this bottom node to it's corresponding top node
    // note: it's not enough to flood once and find best overall route, as
    // only routes starting and ending in the same node are valid
    LOG(4, "flood graph ...");
    double dfWeightMin = 1e10;
    int nNodeMin = 0;
    for (int nNodeCnt = 0; nNodeCnt < aPecPoint[0].size() + 1; nNodeCnt++)
        {
        if (pProgress != 0)
            if (pProgress->setValue((int)(dfProgressMax * (dfProgressFloodBeg +
                (dfProgressFloodEnd - dfProgressFloodBeg) * (double)nNodeCnt / (double)aPecPoint[0].size()))))
                return false;
        // trick: loop over all potential start nodes to find best one;
        // then loop one more time to have graph weights valid for best node
        int nNodeStart = (nNodeCnt < aPecPoint[0].size())? nNodeCnt: nNodeMin;
        // only consider bottom nodes as startpoint that have a
        // corresponding top node which can be reached by a valid route
        if (!aPecPoint[c_nRowNum].at(nNodeStart).linked())
            continue;
        for (int nNodeThis = 0; nNodeThis < aPecPoint[0].size(); nNodeThis++)
            aPecPoint[0][nNodeThis].initWeight(nNodeThis == nNodeStart);
        for (int nLevelThis = 1; nLevelThis < c_nRowNum + 1; nLevelThis++)
            for (int nNodeThis = 0; nNodeThis < aPecPoint[nLevelThis].size(); nNodeThis++)
                aPecPoint[nLevelThis][nNodeThis].calcWeight();
        LOG(5, "    node %d (pos %d, spd %d): %f", nNodeCnt,
            aPecPoint[c_nRowNum].at(nNodeStart).pos(),
            aPecPoint[c_nRowNum].at(nNodeStart).spd(),
            aPecPoint[c_nRowNum].at(nNodeStart).weight());
        if (dfWeightMin > aPecPoint[c_nRowNum].at(nNodeStart).weight())
            {
            dfWeightMin = aPecPoint[c_nRowNum].at(nNodeStart).weight();
            nNodeMin = nNodeStart;
            }
        }
    // fill PEC table according to best route through graph.
    // at first, initialize PEC table with all resets in case of inconsistencies,
    // such as no route through graph at all ...
    LOG(4, "create PEC table ...");
    if (pProgress != 0)
        if (pProgress->setValue((int)(dfProgressMax * dfProgressPecBeg)))
            return false;
    int nPosMin = 1000;
    int nPosMax = -1000;
    m_nFactor = 1;
    for (int nRow = 0; nRow < c_nRowNum; nRow++)
        m_aTable[nRow] = PecSample::reset;
    if (nNodeMin < aPecPoint[c_nRowNum].size())
        {
        const PecNode *pNode = &(aPecPoint[c_nRowNum].at(nNodeMin));
        for (;;)
            {
            const PecNode *pNodePrev = pNode->bestPrev();
            if (pNodePrev == 0)
                break;
            // The firmware seems to model a PEC table entry not as a point,
            // but as an interval. The interval covers the area of the worm
            // where pecByte*4+pecBit is constant.
            // The PEC table entry always needs to be placed at the left side
            // of the intervall, so we need to distinguish traveling direction here
            int nRow = (m_eTraversal == forward)? pNode->row():
                ((pNode->row() - 1 + c_nRowNum) % c_nRowNum);
            int nPos = pNode->pos();
            int nSpd = pNode->spd();
            int nAcc = nSpd - pNodePrev->spd();
            if (nSpd == 0)
                m_aTable[nRow] = PecSample::reset;
            else
                m_aTable[nRow] = PecSample::acc2Val(nAcc);
            if (m_nFactor < abs(nSpd))
                m_nFactor = abs(nSpd);
            if (nPosMin > nPos)
                nPosMin = nPos;
            if (nPosMax < nPos)
                nPosMax = nPos;
            pNode = pNodePrev;
            }
        }
    m_nOffset = (nPosMax < nPosMin)? 0: ((nPosMax + nPosMin) / 2);
    if (pProgress != 0)
        pProgress->setValue((int)(dfProgressMax * dfProgressPecEnd));
#endif

    // Fill in speed and position profile from the generated acceleration sequence
    recalcGeo();

    return true;
    }


/*!
    Recalculate acceleration, speed and position from
    raw PEC table values
*/

void PecData::recalcGeo()
    {
    // traverse twice through the table:
    // - first time to get a correct start speed
    //   necessary for tables without any PecSample::reset value;
    //   still this might not work out, consider e.g. a table of 0's only
    // - second time for storing acc, spd and pos
    int nSpd = 0;
    for (int nRound = 0; nRound < 2; nRound++)
        {
        int nPos = 0;
        for (int nCnt = 0; nCnt < c_nRowNum; nCnt++)
            {
            // careful:
            // The firmware seems to model a PEC table entry not as a point,
            // but as an interval. The interval covers the area of the worm
            // where pecByte*4+pecBit is constant. Physical properties are
            // modeled like this:
            // - acceleration is a spike at the interval entry point, i.e.
            //   it's real position on the worm depends on traversal direction
            // - speed changes immediately at interval entry, since accel is a spike
            //   so speed is the same over the interval, independently of traversal
            //   direction, except if previous speed differs
            // - position reaches resulting value when exiting the interval,
            //   after moving from current to next point with current speed
            // So when looking at a PEC table entry point, it's values depend
            // on traversal direction:
            // - forward (this=n, prev=n-1, next=n+1):
            //   speed[this] = speed[prev] + acc[this]
            //   pos[this] = pos[prev] + speed[prev]
            // - backward (this=n, prev=n+1, next=n-1):
            //   speed[this] = speed[prev] + acc[next]
            //   pos[this] = pos[prev] + speed[prev]

            // nSpd and nPos are speed and position at entry of current interval
            int nRowThis = (m_eTraversal * nCnt + c_nRowNum) % c_nRowNum;
            int nRowNext = (m_eTraversal * (nCnt + 1) + c_nRowNum * 2) % c_nRowNum;
            int nRowAcc = (m_eTraversal == forward)? nRowThis: nRowNext;
            int nAcc = 0;
            switch (m_aTable[nRowAcc].m_eVal)
                {
                case PecSample::reset:  nAcc =  0; nSpd = 0; break;
                case PecSample::faster: nAcc =  1; break;
                case PecSample::slower: nAcc = -1; break;
                case PecSample::keep:   nAcc =  0; break;
                }
            nSpd += nAcc;
            if (nSpd < - m_nFactor)
                nSpd = - m_nFactor;
            if (nSpd > m_nFactor)
                nSpd = m_nFactor;
            if (nRound == 0)
                continue;
            PecSample &rThis = m_aTable[nRowThis];
            rThis.m_nAcc = nAcc;
            rThis.m_nSpd = nSpd;
            rThis.m_nPos = nPos;
            nPos += nSpd;
            }
        }

    // center positions such that m_nOffset is at half distance between minimum and maximum
    int nRowMin = 0;
    int nRowMax = 0;
    for (int nRow = 0; nRow < c_nRowNum; nRow++)
        {
        if (m_aTable.at(nRowMin).m_nPos > m_aTable.at(nRow).m_nPos)
            nRowMin = nRow;
        if (m_aTable.at(nRowMax).m_nPos < m_aTable.at(nRow).m_nPos)
            nRowMax = nRow;
        }
    int nPosMid = (m_aTable.at(nRowMin).m_nPos + m_aTable.at(nRowMax).m_nPos) / 2 - m_nOffset;
    for (int nRow = 0; nRow < c_nRowNum; nRow++)
        m_aTable[nRow].m_nPos -= nPosMid;
    }

