
/***************************************************************************
 *   This file is part of Aspect, a simple PEC tool.                       *
 *                                                                         *
 *   Copyright (C) 2006-2007 by Wolfgang Hoffmann <woho@woho.de>           *
 *                                                                         *
 *   This program is free software, licensed under the GPL v2.             *
 *   See the file COPYING for more details.                                *
 ***************************************************************************/


#include "mainwindow.h"
#include "ui_mainwindow.h"
#include "data.h"
#include "log.h"
#include "mcucmd.h"
#include "mculog.h"
#include "mcuserial.h"

#include <QtGui>
#include <math.h>

/*
#define DEBUG_WOHO
*/




/********************************************************************/
/*!
    \class Progress
    \internal
*/

class Progress : public IProgress
    {
public:
    Progress(QProgressDialog &rDialog)
        : m_pDialog(&rDialog)
        {
        };
    int maxValue() const
        {
        return m_pDialog->maximum();
        };
    bool setValue(int nValue)
        {
        m_pDialog->setValue(nValue);
        qApp->processEvents();
        return m_pDialog->wasCanceled();
        };
private:
    QProgressDialog *m_pDialog;
    };




/********************************************************************/
/*!
    \class MainWindow
    \brief Unsorted heap of all kind of stuff

    MainWindow is the GUI main window.

    \todo sort out stuff that is not related to main window
        (pe and pec handling and calculations, graphs, ...)
    \todo implement residual PE graph
*/

MainWindow::MainWindow()
:   QMainWindow()
    {
    m_pUi = new Ui::MainWindow;
    m_pUi->setupUi(this);

#ifdef DEBUG_WOHO
    m_oCurrentDirectory = QDir(QString::fromUtf8("/home/woho/value/devel/astro/aspect/doc/"));
#else
    m_oCurrentDirectory = QDir::home();
#endif

    m_pMcuLog = new McuLogDialog();

    m_pPeRaw = 0;
    m_aPeProcessed.resize(PecData::c_nRowNum);

    m_pPeProcessedModel = new QStandardItemModel(this);
    m_pUi->viewPeProcessed->setModel(m_pPeProcessedModel);
    m_pPecModel = new QStandardItemModel(this);
    m_pUi->viewPec->setModel(m_pPecModel);
    connect(m_pPecModel, SIGNAL(itemChanged(QStandardItem *)), this, SLOT(changePecItem(QStandardItem *)));
    m_pPecProfileModel = new QStandardItemModel(this);
    m_pUi->viewPecProfile->setModel(m_pPecProfileModel);
    m_pMcuTimer = new QTimer(this);
    connect(m_pMcuTimer, SIGNAL(timeout()), this, SLOT(mcuTimer()));

    m_pMcuPort = new McuSerial;
    connect(m_pMcuPort, SIGNAL(log(const QString &)), m_pMcuLog, SLOT(append(const QString &)));

    m_pPeChart = new QGraphicsScene(this);
    m_pUi->chartPe->setScene(m_pPeChart);

    // todo: implement functionality
    m_pUi->toggleChartPeResidual->setChecked(false);
    m_pUi->toggleChartPeResidual->setEnabled(false);

    m_dfSiderealDay = 86164.091;
    calcPecParams(false);

    m_eGuideDir = stop;
    m_pUi->mcuPause->setEnabled(false);

/*
#ifdef DEBUG_WOHO
#else
    m_pUi->tabPecProfile->setEnabled(false);
    m_pUi->toggleChartPecProfile->setChecked(false);
    m_pUi->toggleChartPecProfile->setEnabled(false);
#endif
*/

#ifdef DEBUG_WOHO
    // woho's personal defaults to ease testing :-)
    // pixel scale: measured: 1.37, calculated: 1.31 (5.6µm @ 880mm: atan2(5.6e-6, 880e-3) / M_PI * 180. * 3600.)
    m_pUi->widgetPeIris->setPixelScale(1.31, 15. + 14./60. + 38./3600.);
    m_pUi->widgetPeIris->setWormSync(55., 91);
#if 0
    m_pUi->peProcDrift->setValue(-1.14);
    openPeFile(QString::fromUtf8("/media/sdb8/pub/webcam/pe4x.dat"));
#elif 1
    m_pUi->peProcDrift->setValue(0.);
    openPeFile(QString::fromUtf8("/home/woho/value/devel/astro/aspect/doc/pe_woho/woho_pec03.csv"));
#else
    m_pUi->peProcDrift->setValue(.3);
    openPeFile(QString::fromUtf8("/home/woho/value/devel/astro/aspect/doc/doug_dieter/pec_bad_drift.csv"));
#endif
    openPecFile(QString::fromUtf8("/media/sdb8/pub/webcam/littlefoot/pec_20061015a.csv"));
#endif

    resize(960, 640);
    }


MainWindow::~MainWindow()
    {
    delete m_pMcuLog;
    delete m_pUi;
    delete m_pMcuPort;
    }


void MainWindow::on_actionQuit_triggered()
    {
    LOG(4, "MainWindow::on_actionQuit_triggered");
    close();
    }


void MainWindow::on_actionAbout_Aspect_triggered()
    {
    LOG(4, "MainWindow::on_actionAbout_Aspect_triggered");
    QMessageBox::about(this,
        tr("About Aspect ..."),
        tr("<h3>Aspect</h3>"
           "<p>Aspect is a simple PEC tool, allowing to "
           "view, edit and calculate periodic error correction "
           "tables for the <a href=\"http://littlefoot.rajiva.de/\">LittleFoot</a> "
           "astronomical mount stepper controler.</p>"
           "<p>Version: <b>0.7</b> (04 Feb 2007)<br>"
           "Autor: Wolfgang Hoffmann (<a href=\"mailto:woho@woho.de\">woho@woho.de</a>)<br>"
           "Homepage: <a href=\"http://www.woho.de/devel/astro/aspect.html\">"
           "http://www.woho.de/devel/astro/aspect.html</a></p>"
           "<p>Copyright (C) 2006-2007 Wolfgang Hoffmann.<br>"
           "This program is free software, licensed under the GPL v2.</p>"
           "<p>This program is distributed in the hope that it will be useful, "
           "but WITHOUT ANY WARRANTY; without even the implied warranty of "
           "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the "
           "GNU General Public License for more details.</p>"));
    }


void MainWindow::on_actionAbout_Qt_triggered()
    {
    LOG(4, "MainWindow::on_actionAbout_Qt_triggered");
    QApplication::aboutQt();
    }


void MainWindow::on_actionMCU_Log_View_triggered()
    {
    m_pMcuLog->show();
    }




/********************************************************************/
/*
    MCU connection
*/


void MainWindow::on_mcuPort_activated(const QString &qsPort)
    {
    Q_UNUSED(qsPort)
    }


void MainWindow::on_mcuAutoconnect_toggled(bool bOn)
    {
    if (!bOn)
        m_pUi->mcuAutoconnect->setChecked(true);
    }


/*!
    \todo implement
*/

void MainWindow::on_mcuConnect_toggled(bool bOn)
    {
    LOG(0, "on_mcuConnect_toggled(%s)", bOn? "on": "off");
    }


void MainWindow::on_mcuSettingsTracking_activated(int nSpeed)
    {
    McuDataTracking oTracking;
    oTracking.setRate(
        (nSpeed == 0)? McuDataTracking::sidereal:
        (nSpeed == 1)? McuDataTracking::solar:
        (nSpeed == 2)? McuDataTracking::lunar:
        McuDataTracking::off);
    if (mcuCommunicationBegin())
        {
        oTracking.mcuSet(*m_pMcuPort);
        mcuCommunicationEnd();
        }
    }


void MainWindow::on_mcuSettingsPec_activated(int nIndex)
    {
    if (m_pUi->mcuSettingsPec->count() == 1)
        return;
    McuDataPecState oPec;
    oPec.setPec(nIndex == 1);
    if (mcuCommunicationBegin())
        {
        oPec.mcuSetPec(*m_pMcuPort);
        mcuCommunicationEnd();
        }
    }


void MainWindow::on_mcuSettingsIntelly_activated(int nIndex)
    {
    McuDataPecState oPec;
    oPec.setIntellyTrack(nIndex == 1);
    if (mcuCommunicationBegin())
        {
        oPec.mcuSetIntelly(*m_pMcuPort);
        mcuCommunicationEnd();
        }
    }


void MainWindow::on_mcuSettingsRefresh_clicked()
    {
    McuDataState oState;
    McuDataPecState oPec;
    if (mcuCommunicationBegin())
        {
        oState.mcuGet(*m_pMcuPort);
        oPec.mcuGet(*m_pMcuPort);
        mcuCommunicationEnd();
        }

    m_pUi->mcuSettingsTracking->setCurrentIndex(oState.trackingRate());

    // dynamically change contents of PEC dropdown, as single entries cannot be disabled
    if (oPec.pec() < 2)
        {
        m_pUi->mcuSettingsPec->setItemText(0, tr("Off"));
        if (m_pUi->mcuSettingsPec->count() == 1)
            m_pUi->mcuSettingsPec->addItem(tr("On"));
        else
            m_pUi->mcuSettingsPec->setItemText(1, tr("On"));
        m_pUi->mcuSettingsPec->setCurrentIndex(oPec.pec());
        }
    else
        {
        if (m_pUi->mcuSettingsPec->count() == 2)
            m_pUi->mcuSettingsPec->removeItem(1);
        if (oPec.pec() == 2)
            m_pUi->mcuSettingsPec->setItemText(0, tr("On by Handbox"));
        else
            m_pUi->mcuSettingsPec->setItemText(0, tr("In Training"));
        }

    m_pUi->mcuSettingsIntelly->setCurrentIndex(oPec.intellyTrack()? 1: 0);
    m_pUi->mcuSettingsRates->setText(
        QString::fromUtf8("2x: %1, 8x: %2, 16x: %3, Goto: %4")
        .arg(oState.guideRate())
        .arg(oState.paddle8Rate())
        .arg(oState.paddle16Rate())
        .arg(oState.gotoRate()));
    m_pUi->mcuSettingsMount->setText(oState.rotOrient());
    m_pUi->mcuSettingsGears->setText(
        QString::fromUtf8("Gear Ratio: %1, Worm Ratio: %2, Worm Wheel Teeth: %3")
        .arg(oState.gearboxReduction())
        .arg(oState.wormRatio())
        .arg(oState.wormTeeth()));
    m_pUi->mcuSettingsBacklash->setText(oState.backlash());
    }


void MainWindow::on_mcuCommandSend_clicked()
    {
    LOG(4, "MainWindow::on_mcuCommand_editingFinished");
    QByteArray aCommand = m_pUi->mcuCommand->currentText().toLatin1().constData();
    McuRawCommand oCmd;
    oCmd.setCommand(aCommand);
    if (mcuCommunicationBegin())
        {
        oCmd.mcuSend(*m_pMcuPort);
        mcuCommunicationEnd();
        }
    QByteArray aAnswer = oCmd.answer();
    QString qsAnswer;
    for (int nByte = 0; nByte < aAnswer.length(); nByte++)
        qsAnswer += QString::fromUtf8("%1 ")
            .arg((int)((unsigned char)aAnswer.at(nByte)), 2, 16, QLatin1Char('0'));
    qsAnswer += '"';
    for (int nByte = 0; nByte < aAnswer.length(); nByte++)
        {
        QChar oChar = QChar::fromLatin1(aAnswer.at(nByte));
        qsAnswer += (oChar.isPrint())? oChar: QLatin1Char('.');
        }
    qsAnswer += '"';
    m_pUi->mcuAnswer->setText(qsAnswer);
    }


/*!
    \todo modularize enabling/disabling of MCU communication related buttons
        (use signals or list of interfaces)
*/

bool MainWindow::mcuCommunicationBegin()
    {
    m_pUi->mcuSettingsTracking->setEnabled(false);
    m_pUi->mcuSettingsPec->setEnabled(false);
    m_pUi->mcuSettingsIntelly->setEnabled(false);
    m_pUi->mcuSettingsRefresh->setEnabled(false);
    m_pUi->mcuCommandSend->setEnabled(false);
    m_pUi->pecGetTable->setEnabled(false);
    m_pUi->pecPutTable->setEnabled(false);
    m_pUi->pecMountQuery->setEnabled(false);
    m_pUi->mcuGuideEast->setEnabled(false);
    m_pUi->mcuGuideWest->setEnabled(false);
    if (!m_pUi->mcuStart->isChecked())
        m_pUi->mcuStart->setEnabled(false);
    if (m_pUi->mcuPause->isChecked())
        m_pUi->mcuPause->setEnabled(false);

    if (m_pMcuPort->open(m_pUi->mcuPort->currentText()))
        return true;

    m_pUi->mcuSettingsTracking->setEnabled(true);
    m_pUi->mcuSettingsPec->setEnabled(true);
    m_pUi->mcuSettingsIntelly->setEnabled(true);
    m_pUi->mcuSettingsRefresh->setEnabled(true);
    m_pUi->mcuCommandSend->setEnabled(true);
    m_pUi->pecGetTable->setEnabled(true);
    m_pUi->pecPutTable->setEnabled(true);
    m_pUi->pecMountQuery->setEnabled(true);
    m_pUi->mcuGuideEast->setEnabled(true);
    m_pUi->mcuGuideWest->setEnabled(true);
    if (!m_pUi->mcuStart->isChecked())
        m_pUi->mcuStart->setEnabled(true);
    if (m_pUi->mcuPause->isChecked())
        m_pUi->mcuPause->setEnabled(true);
    return false;
    }


void MainWindow::mcuCommunicationEnd()
    {
    m_pMcuPort->close();

    m_pUi->mcuSettingsTracking->setEnabled(true);
    m_pUi->mcuSettingsPec->setEnabled(true);
    m_pUi->mcuSettingsIntelly->setEnabled(true);
    m_pUi->mcuSettingsRefresh->setEnabled(true);
    m_pUi->mcuCommandSend->setEnabled(true);
    m_pUi->pecGetTable->setEnabled(true);
    m_pUi->pecPutTable->setEnabled(true);
    m_pUi->pecMountQuery->setEnabled(true);
    m_pUi->mcuGuideEast->setEnabled(true);
    m_pUi->mcuGuideWest->setEnabled(true);
    if (!m_pUi->mcuStart->isChecked())
        m_pUi->mcuStart->setEnabled(true);
    if (m_pUi->mcuPause->isChecked())
        m_pUi->mcuPause->setEnabled(true);
    }




/********************************************************************/
/*
    PE raw data
*/


void MainWindow::on_actionOpen_PE_Data_triggered()
    {
    LOG(4, "MainWindow::on_actionOpen_PE_Data_triggered");

    // switch off polling
    m_pUi->peFileFollow->setChecked(false);

    QFileDialog oFileDialog(this);
    oFileDialog.setAcceptMode(QFileDialog::AcceptOpen);
    oFileDialog.setFileMode(QFileDialog::ExistingFile);
    oFileDialog.setViewMode(QFileDialog::Detail);
    oFileDialog.setWindowTitle(tr("Open PE Data File"));
    oFileDialog.setDirectory(m_oCurrentDirectory);
    QStringList qslFilters;
    qslFilters << tr("MCU Control Plus Format (*.csv)");
    qslFilters << tr("IRIS Format (*.dat)");
    qslFilters << tr("All Files (*)");
    oFileDialog.setFilters(qslFilters);

    QStringList qslFileNames;
    if (oFileDialog.exec())
        {
        m_oCurrentDirectory = oFileDialog.directory();
        qslFileNames = oFileDialog.selectedFiles();
        }
    if (qslFileNames.size() > 0)
        {
        QString qsFileName = qslFileNames.at(0);
        LOG(3, "%s", qsFileName.toLocal8Bit().constData());
        openPeFile(qsFileName);
        }
    }


void MainWindow::on_peFileFollow_toggled(bool bOn)
    {
    LOG(0, "MainWindow::on_peFileFollow_toggled");
    if (m_pPeRaw == 0)
        return;
    if (!m_pPeRaw->pollPeFile(bOn))
        m_pUi->peFileFollow->setChecked(false);
    }


void MainWindow::openPeFile(const QString &qsFileName)
    {
    LOG(4, "MainWindow::openPeFile");
    m_pUi->lineEditPeFile->setText(qsFileName);

    if (m_pPeRaw != 0)
        m_pPeRaw->clearData();

    if (m_pUi->widgetPeMcu->probePeFile(qsFileName))
        {
        m_pUi->stackPeFile->setCurrentWidget(m_pUi->pagePeMcu);
        m_pPeRaw = m_pUi->widgetPeMcu;
        }
    else if (m_pUi->widgetPeIris->probePeFile(qsFileName))
        {
        m_pUi->stackPeFile->setCurrentWidget(m_pUi->pagePeIris);
        m_pPeRaw = m_pUi->widgetPeIris;
        }
    else
        {
        m_pUi->stackPeFile->setCurrentWidget(m_pUi->pagePeNone);
        m_pPeRaw = 0;
        }

    if (m_pPeRaw != 0)
        {
        // will emit dataPeChanged() connected to recalcPeData()
        m_pPeRaw->setPeFileName(qsFileName);
        if (m_pPeRaw->loadPeFile())
            m_pUi->peFileFollow->setEnabled(true);
        }
    else
        {
        m_pUi->peFileFollow->setEnabled(false);
        m_pUi->peFileFollow->setChecked(false);
        recalcPeData();
        }
    }


void MainWindow::on_widgetPeMcu_dataPeChanged(IPeRawData *pData)
    {
    LOG(4, "MainWindow::on_widgetPeMcu_dataPeChanged");
    if (m_pPeRaw == pData)
        recalcPeData();
    }


void MainWindow::on_widgetPeIris_dataPeChanged(IPeRawData *pData)
    {
    LOG(4, "MainWindow::on_widgetPeIris_dataPeChanged");
    if (m_pPeRaw == pData)
        recalcPeData();
    }




/********************************************************************/
/*
    PE processing
*/


void MainWindow::on_peProcMode_activated(int nMode)
    {
    LOG(4, "MainWindow::on_peProcMode_activated");
    Q_UNUSED(nMode)
    recalcPeData();
    }


void MainWindow::on_peProcBeg_valueChanged(int nBeg)
    {
    LOG(4, "MainWindow::on_peProcBeg_valueChanged");
    Q_UNUSED(nBeg)
    recalcPeData();
    }


void MainWindow::on_peProcEnd_valueChanged(int nEnd)
    {
    LOG(4, "MainWindow::on_peProcEnd_valueChanged");
    Q_UNUSED(nEnd)
    recalcPeData();
    }


void MainWindow::on_peProcCombining_activated(int nMethod)
    {
    LOG(4, "MainWindow::on_peProcCombining_activated");
    Q_UNUSED(nMethod)
    recalcPeData();
    }


void MainWindow::on_peProcSmoothWidth_valueChanged(int nWidth)
    {
    LOG(4, "MainWindow::on_peProcSmoothWidth_valueChanged");
    Q_UNUSED(nWidth)
    recalcPeData();
    }


void MainWindow::on_peProcDrift_valueChanged(double dfDrift)
    {
    LOG(4, "MainWindow::on_peProcDrift_valueChanged");
    Q_UNUSED(dfDrift)
    recalcPeData();
    }


void MainWindow::on_peProcDecl_valueChanged(double dfDeclination)
    {
    LOG(4, "MainWindow::on_peProcDecl_valueChanged");
    Q_UNUSED(dfDeclination)
    recalcPeData();
    }


void MainWindow::on_peProcScale_valueChanged(double dfScale)
    {
    LOG(4, "MainWindow::on_peProcScale_valueChanged");
    Q_UNUSED(dfScale)
    recalcPeData();
    }


void MainWindow::on_peProcOffset_valueChanged(double dfOffset)
    {
    LOG(4, "MainWindow::on_peProcOffset_valueChanged");
    Q_UNUSED(dfOffset)
    recalcPeData();
    }


void MainWindow::on_toggleChartPeRaw_toggled(bool bOn)
    {
    LOG(4, "MainWindow::on_toggleChartPeRaw_toggled");
    Q_UNUSED(bOn)
    recalcPeChart();
    }


void MainWindow::on_toggleChartPeProcessed_toggled(bool bOn)
    {
    LOG(4, "MainWindow::on_toggleChartPeProcessed_toggled");
    Q_UNUSED(bOn)
    recalcPeChart();
    }


void MainWindow::on_toggleChartPec_toggled(bool bOn)
    {
    LOG(4, "MainWindow::on_toggleChartPec_toggled");
    Q_UNUSED(bOn)
    recalcPeChart();
    }


void MainWindow::on_toggleChartPecProfile_toggled(bool bOn)
    {
    LOG(4, "MainWindow::on_toggleChartPecProfile_toggled");
    Q_UNUSED(bOn)
    recalcPeChart();
    }


void MainWindow::on_toggleChartPeResidual_toggled(bool bOn)
    {
    LOG(4, "MainWindow::on_toggleChartPeResidual_toggled");
    Q_UNUSED(bOn)
    recalcPeChart();
    }




/********************************************************************/
/*
    PEC table
*/


void MainWindow::on_actionOpen_PEC_Table_triggered()
    {
    LOG(4, "MainWindow::on_actionOpen_PEC_Table_triggered");
    QFileDialog oFileDialog(this);
    oFileDialog.setAcceptMode(QFileDialog::AcceptOpen);
    oFileDialog.setFileMode(QFileDialog::ExistingFile);
    oFileDialog.setViewMode(QFileDialog::Detail);
    oFileDialog.setWindowTitle(tr("Open PEC Table File"));
    oFileDialog.setDirectory(m_oCurrentDirectory);
    QStringList qslFilters;
    qslFilters << tr("PEC File (*.csv)");
    qslFilters << tr("All Files (*)");
    oFileDialog.setFilters(qslFilters);

    QStringList qslFileNames;
    if (oFileDialog.exec())
        {
        m_oCurrentDirectory = oFileDialog.directory();
        qslFileNames = oFileDialog.selectedFiles();
        }
    if (qslFileNames.size() > 0)
        {
        QString qsFileName = qslFileNames.at(0);
        LOG(3, "%s", qsFileName.toLocal8Bit().constData());
        openPecFile(qsFileName);
        }
    }


void MainWindow::openPecFile(const QString &qsFileName)
    {
    LOG(4, "MainWindow::openPecFile");

    QFile oFile(qsFileName);
    oFile.open(QIODevice::ReadOnly | QIODevice::Text);
    QTextStream oStream(&oFile);
    QString qsHeader = oStream.readLine();
    if (qsHeader != QString::fromUtf8("Factor, Table, ReqSpeed, IsSpeed, MinSpeed, MaxSpeed"))
        {
        LOG(0, "MainWindow::openPecFile(): wrong file format");
        return;
        }
    m_oPecTable.clear();
    for (int nRow = 0; nRow < PecData::c_nRowNum; nRow++)
        {
        if (oStream.atEnd())
            continue;
        QString qsLine = oStream.readLine();
        LOG(5, "    %s", qsLine.toLocal8Bit().constData());
        QStringList qslEntryStrings = qsLine.split(",");
        int nFactor = (qslEntryStrings.size() > 0)? qslEntryStrings.at(0).toInt(): 1;
        int nTable = (qslEntryStrings.size() > 1)? qslEntryStrings.at(1).toInt(): -1;
        //int nReqSpeed = (qslEntryStrings.size() > 2)? qslEntryStrings.at(2).toInt(): -1;
        //int nIsSpeed = (qslEntryStrings.size() > 3)? qslEntryStrings.at(3).toInt(): -1;
        //int nMinSpeed = (qslEntryStrings.size() > 4)? qslEntryStrings.at(4).toInt(): -1;
        //int nMaxSpeed = (qslEntryStrings.size() > 5)? qslEntryStrings.at(5).toInt(): -1;
        if (nRow == 0)
            m_oPecTable.setFactor(nFactor);
        m_oPecTable.setValue(nRow, PecSample::int2Val(nTable));
        }
    m_oPecTable.setFinish();
    oFile.close();

    fillPecModel();
    recalcPeChart();
    }


void MainWindow::on_actionSave_PEC_Table_triggered()
    {
    LOG(4, "MainWindow::on_actionSave_PEC_Table_triggered");
    QFileDialog oFileDialog(this);
    oFileDialog.setAcceptMode(QFileDialog::AcceptSave);
    oFileDialog.setFileMode(QFileDialog::AnyFile);
    oFileDialog.setViewMode(QFileDialog::Detail);
    oFileDialog.setWindowTitle(tr("Save PEC Table As File"));
    oFileDialog.setDirectory(m_oCurrentDirectory);
    QStringList qslFilters;
    qslFilters << tr("PEC File (*.csv)");
    qslFilters << tr("All Files (*)");
    oFileDialog.setFilters(qslFilters);
    oFileDialog.setDefaultSuffix(QString::fromUtf8("csv"));

    QStringList qslFileNames;
    if (oFileDialog.exec())
        {
        m_oCurrentDirectory = oFileDialog.directory();
        qslFileNames = oFileDialog.selectedFiles();
        }
    if (qslFileNames.size() > 0)
        {
        QString qsFileName = qslFileNames.at(0);
        LOG(3, "%s", qsFileName.toLocal8Bit().constData());
        savePecFile(qsFileName);
        }
    }


void MainWindow::savePecFile(const QString &qsFileName)
    {
    LOG(4, "MainWindow::savePecFile");

    QFile oFile(qsFileName);
    // don't use flag QIODevice::Text here, because we need to write CR/LF on all platforms
    oFile.open(QIODevice::WriteOnly);
    QTextStream oStream(&oFile);
    oStream << QString::fromUtf8("Factor, Table, ReqSpeed, IsSpeed, MinSpeed, MaxSpeed\r\n");
    int nReqSpeed = 0;
    for (int nRow = 0; nRow < PecData::c_nRowNum; nRow++)
        {
        int nFactor = m_oPecTable.factor();
        int nTable = PecSample::val2Int(m_oPecTable[nRow].m_eVal);
        if (nTable == 1)
            nReqSpeed += nFactor;
        else if (nTable == 2)
            nReqSpeed -= nFactor;
        else if (nTable == 3)
            nReqSpeed = 0;
        int nIsSpeed = nReqSpeed;
        int nMinSpeed = -47;
        int nMaxSpeed = 47;
        if (nIsSpeed < nMinSpeed)
            nIsSpeed = nMinSpeed;
        if (nIsSpeed > nMaxSpeed)
            nIsSpeed = nMaxSpeed;
        // careful:
        // write CR LF to make MCU Update Plus cope with our PEC files
        oStream << QString::fromUtf8("%1, %2, %3, %4, %5, %6\r\n")
            .arg(nFactor).arg(nTable)
            .arg(nReqSpeed).arg(nIsSpeed)
            .arg(nMinSpeed).arg(nMaxSpeed);
        }
    oStream << QString::fromUtf8("\r\n");
    oFile.close();
    }


void MainWindow::on_pecGetTable_clicked()
    {
    LOG(4, "MainWindow::on_pecGetTable_clicked");
    getPecTable();
    }


/*!
    Download PEC table from MCU via serial line

    \todo fix hang if port is not available
    \todo make port configurable
    \todo evaluate return values and do some sane error handling here ...
*/

void MainWindow::getPecTable()
    {
    LOG(4, "MainWindow::getPecTable");

    QProgressDialog oProgress(tr("Downloading PEC table from MCU ..."),
        tr("Cancel"), 0, PecData::c_nRowNum, this);
    oProgress.setWindowTitle(tr("Aspect - Get PEC table"));
    oProgress.setMinimumDuration(0);
    oProgress.setWindowModality(Qt::WindowModal);
    QApplication::setOverrideCursor(Qt::BusyCursor);

    McuDataPecTable oPec;
    if (mcuCommunicationBegin())
        {
        oPec.mcuGetFactor(*m_pMcuPort);
        oPec.mcuGetValuesBegin(*m_pMcuPort);
        for (;;)
            {
            int nRow = oPec.mcuGetValues(*m_pMcuPort);
            if (nRow >= 0)
                oProgress.setValue(nRow);
            qApp->processEvents();

            if (nRow == PecData::c_nRowNum)
                break;
            else if ((nRow == -1) || (oProgress.wasCanceled()))
                {
                oPec.mcuGetValuesAbort(*m_pMcuPort);
                break;
                }
            }
        mcuCommunicationEnd();
        }
    m_oPecTable = oPec.data(m_oPecTable.traversal());
    QApplication::restoreOverrideCursor();

    fillPecModel();
    recalcPeChart();
    }


void MainWindow::on_pecPutTable_clicked()
    {
    LOG(4, "MainWindow::on_pecPutTable_clicked");
    putPecTable();
    }


/*!
    Upload PEC table to MCU via serial line

    \todo show message box if opening port failed
    \todo evaluate return values and do some sane error handling here ...
    \todo query and display PEC status after uploading PEC table
    \todo verify if writing order is ok (factor, save to EEPROM, PEC values, save PEC)
*/

void MainWindow::putPecTable()
    {
    LOG(4, "MainWindow::putPecTable");

    QProgressDialog oProgress(tr("Uploading PEC table to MCU ..."),
        tr("Cancel"), 0, PecData::c_nRowNum + 2, this);
    oProgress.setWindowTitle(tr("Aspect - Put PEC table"));
    oProgress.setMinimumDuration(0);
    oProgress.setWindowModality(Qt::WindowModal);
    QApplication::setOverrideCursor(Qt::BusyCursor);

    McuDataPecTable oPec;
    oPec.set(m_oPecTable);

    if (mcuCommunicationBegin())
        {
        oProgress.setValue(0);
        oPec.mcuSetFactor(*m_pMcuPort);
        for (int nRow = 0; nRow < PecData::c_nRowNum; nRow++)
            {
            oProgress.setValue(nRow + 1);
            qApp->processEvents();

            if (oProgress.wasCanceled())
                break;
            oPec.mcuSetValue(*m_pMcuPort, nRow);
            }
        oProgress.setValue(PecData::c_nRowNum + 1);
        oPec.mcuSave(*m_pMcuPort);
        oProgress.setValue(PecData::c_nRowNum + 2);
        mcuCommunicationEnd();
        }
    QApplication::restoreOverrideCursor();
    }


void MainWindow::fillPecModel()
    {
    m_pPecModel->clear();
    m_pPecModel->setColumnCount(6);
    m_pPecModel->setRowCount(PecData::c_nRowNum);

    QStringList qslHeaders;
    qslHeaders << tr("Worm Position");
    qslHeaders << tr("PEC");
    qslHeaders << tr("acc");
    qslHeaders << tr("spd");
    qslHeaders << tr("pos");
    qslHeaders << tr("");
    m_pPecModel->setHorizontalHeaderLabels(qslHeaders);
    if (m_pUi->viewPec->header() != 0)
        m_pUi->viewPec->header()->setResizeMode(QHeaderView::ResizeToContents);

    m_pUi->pecFactor->setValue(m_oPecTable.factor());
    for (int nRow = 0; nRow < PecData::c_nRowNum; nRow++)
        {
        QStandardItem *pItem;
        pItem = new QStandardItem(QString::number(nRow));
        pItem->setEditable(false);
        pItem->setSelectable(false);
        m_pPecModel->setItem(nRow, 0, pItem);
        pItem = new QStandardItem(QString::number(m_oPecTable.valueInt(nRow)));
        pItem->setEditable(true);
        pItem->setSelectable(true);
        m_pPecModel->setItem(nRow, 1, pItem);
        pItem = new QStandardItem(QString::number(m_oPecTable[nRow].m_nAcc));
        pItem->setEditable(false);
        pItem->setSelectable(false);
        m_pPecModel->setItem(nRow, 2, pItem);
        pItem = new QStandardItem(QString::number(m_oPecTable[nRow].m_nSpd));
        pItem->setEditable(false);
        pItem->setSelectable(false);
        m_pPecModel->setItem(nRow, 3, pItem);
        pItem = new QStandardItem(QString::number(m_oPecTable[nRow].m_nPos));
        pItem->setEditable(false);
        pItem->setSelectable(false);
        m_pPecModel->setItem(nRow, 4, pItem);
        pItem = new QStandardItem(QString());
        pItem->setEditable(false);
        pItem->setSelectable(false);
        m_pPecModel->setItem(nRow, 5, pItem);
        }
    }


void MainWindow::changePecItem(QStandardItem *pItem)
    {
    LOG(5, "MainWindow::changePecItem");
    if ((pItem == 0) || (pItem->column() != 1))
        return;
    int nRow = pItem->row();
    if ((nRow < 0) || (nRow >= PecData::c_nRowNum))
        return;
    int nPec = pItem->text().toInt();
    if (m_oPecTable.value(nRow) == PecSample::int2Val(nPec))
        return;
    m_oPecTable.setValue(nRow, PecSample::int2Val(nPec));
    m_oPecTable.setFinish();
    recalcPeChart();
    }


void MainWindow::on_pecFactor_valueChanged(int nFactor)
    {
    LOG(5, "MainWindow::on_pecFactor_valueChanged");
    if (m_oPecTable.factor() == nFactor)
        return;

    m_oPecTable.setFactor(nFactor);
    m_oPecTable.setFinish();
    fillPecModel();
    recalcPeChart();
    }


void MainWindow::on_pecGraphOffset_valueChanged(double dfOffset)
    {
    LOG(4, "MainWindow::on_pecGraphOffset_valueChanged");
    Q_UNUSED(dfOffset)
    recalcPeChart();
    }


/*!
    \todo show message box if opening port failed
*/

void MainWindow::on_pecMountQuery_clicked()
    {
    LOG(4, "MainWindow::on_pecMountQuery_clicked");
    McuDataState oState;
    bool bSuccess = false;
    if (mcuCommunicationBegin())
        {
        bSuccess = oState.mcuGet(*m_pMcuPort);
        mcuCommunicationEnd();
        }
    if (!bSuccess)
        {
        LOG(0, "failed to query Worm Wheel Teeth from MCU");
        return;
        }
    int nWormWheelTeeth = oState.wormTeeth();
    if (nWormWheelTeeth <= 0)
        {
        LOG(0, "ignoring values, Worm Wheel Teeth %d looks strange",
            nWormWheelTeeth);
        return;
        }
    m_pUi->pecWormWheel->setValue(nWormWheelTeeth);
    }


void MainWindow::on_pecWormWheel_valueChanged(int nTeeth)
    {
    LOG(4, "MainWindow::on_pecWormWheel_valueChanged");
    Q_UNUSED(nTeeth)
    calcPecParams();
    }


void MainWindow::on_pecTraversal_activated(int nDirection)
    {
    LOG(4, "MainWindow::on_pecTraversal_activated");
    Q_UNUSED(nDirection)
    calcPecParams();
    }


void MainWindow::on_pecAccel_valueChanged(double dfAccel)
    {
    LOG(4, "MainWindow::on_pecAccel_valueChanged");
    Q_UNUSED(dfAccel)
    calcPecParams();
    }


void MainWindow::calcPecParams(bool bRedrawPecChart)
    {
    LOG(4, "MainWindow::calcPecParams");

    m_nWormWheelTeeth = m_pUi->pecWormWheel->value();
    double dfAccel = m_pUi->pecAccel->value();
    double dfArcSecsPerPecPoint = 360. * 3600. / (double)m_nWormWheelTeeth / (double)PecData::c_nRowNum;
    double dfPecPosScale = dfAccel * dfArcSecsPerPecPoint;

    PecData::Traversal eTraversal = (m_pUi->pecTraversal->currentIndex() == 0)?
        PecData::backward: PecData::forward;

    bool bRedraw = bRedrawPecChart &&
        ((m_dfPecPosScale != dfPecPosScale) ||
         (m_oPecTable.traversal() != eTraversal));
    m_dfPecPosScale = dfPecPosScale;
    m_oPecTable.setTraversal(eTraversal);

    m_pUi->mcuStatusTeeth->setText(QString::number(m_nWormWheelTeeth));
    m_pUi->mcuStatusTraversal->setText(
        QString::fromUtf8((eTraversal == PecData::forward)? "forward": "backward"));

    if (bRedraw)
        {
        fillPecModel();
        recalcPeChart();
        }
    }


void MainWindow::on_pecCalculate_clicked()
    {
    LOG(4, "MainWindow::on_pecCalculate_clicked");
    calcPecTable();
    }


/*!
    Calculate PEC table from processed PE data

    \sa PecData::calc()

    \todo make sure that processed PE data is valid
*/

void MainWindow::calcPecTable()
    {
    LOG(4, "MainWindow::calcPecTable");

    QProgressDialog oProgress(tr("Calculating PEC table from PE measurement ..."),
        tr("Cancel"), 0, PecData::c_nRowNum, this);
    oProgress.setWindowTitle(tr("Aspect - Calculate PEC table"));
    oProgress.setMinimumDuration(0);
    oProgress.setWindowModality(Qt::WindowModal);
    QApplication::setOverrideCursor(Qt::BusyCursor);

    // convert processed PE data into a profile for compensating the PE
    QVector<double> aProfile(PecData::c_nRowNum);
    for (int nRow = 0; nRow < PecData::c_nRowNum; nRow++)
        aProfile[nRow] = m_aPeProcessed.at(nRow) / m_dfPecPosScale;
    {
    Progress oPrg(oProgress);
    m_oPecTable.calc(aProfile, &oPrg);
    }

    QApplication::restoreOverrideCursor();

    fillPecModel();
    recalcPeChart();
    }




/********************************************************************/
/*
    ...
*/


/*!
    \todo rework drift compensation
        - it relies on the presence of precise timestamps
            (it timestamp lack msecs, worm pos has better resolution, but is
            proportional to time only if PE measurement was done with PEC and IntellyTrack off)
        - it probably should consider the DEC values and work with something
            more sophisticated than linear RA scaling
*/

void MainWindow::recalcPeData()
    {
    LOG(4, "MainWindow::recalcPeData");

    m_lstPePure.clear();

    double dfDrift = m_pUi->peProcDrift->value() / 60000.;
    double dfScale = m_pUi->peProcScale->value() / cos(m_pUi->peProcDecl->value() * M_PI / 180.);
    double dfOffset = m_pUi->peProcOffset->value();
    QList<PeRawSample> oData = (m_pPeRaw != 0)? m_pPeRaw->samples(): QList<PeRawSample>();
    int nBeg = m_pUi->peProcBeg->value();
    int nEnd = m_pUi->peProcEnd->value();
    if ((nEnd < 0) || (nEnd > oData.size()))
        nEnd = oData.size();
    for (int nRow = nBeg; nRow < nEnd; nRow++)
        {
        PePureSample oSample;
        oSample.m_dfWorm = oData.at(nRow).m_dfWorm;
        double dfTime = oData.at(0).m_oTime.msecsTo(oData.at(nRow).m_oTime);
        oSample.m_dfPe = (oData.at(nRow).m_dfRa - dfTime * dfDrift) * dfScale + dfOffset;
        m_lstPePure.append(oSample);
        LOG(5, "    %.03f: %.03f", oSample.m_dfWorm, oSample.m_dfPe);
        }

    // fist smooth raw sample curves for display
    bool bAverage = (m_pUi->peProcCombining->currentIndex() == 0);  // 0=Average, 1=Median
    int nSmoothWidth = m_pUi->peProcSmoothWidth->value();
    m_lstPeSmoothed.clear();
    if ((nSmoothWidth == 0) || (m_lstPePure.isEmpty()))
        {
        // if without smoothing, pass through original data
        m_lstPeSmoothed = m_lstPePure;
        }
    else
        {
        // combine samples into PEC point grid,
        // taking into account samples of nSmoothWidth PEC points

        // walk over raw data, keeping record of begin, mid and end of
        // smoothing interval:
        // - advance end of interval until hitting the end of a PEC point
        // - advance begin of interval as long as it's further from end
        //   than smoothing width
        // - calculate average/median of data samples within the interval
        //   and store at end index into list of smoothed samples
        int nRowBeg = -1;
        int nRowEnd = 0;
        int nPecMid = -1;
        int nPecBeg = -1;
        int nPecEnd = -1;
        for (int nRowMid = 0; nRowMid < m_lstPePure.size(); nRowMid++)
            {
            int nPecPrev = nPecMid;
            nPecMid = (int)floor(m_lstPePure.at(nRowMid).m_dfWorm);
            if (nPecMid == nPecPrev)
                continue;

            // advance begin of data range while data sample is outside
            // left half of smoothing interval
            // be careful to cope with both PEC table traversal directions
            // and correctly deal with PEC point wraparound and initial case
            for (;;)
                {
                int nPecDiff = abs(nPecMid - nPecBeg);
                if (nPecDiff > PecData::c_nRowNum / 2)
                    nPecDiff = abs(PecData::c_nRowNum - nPecDiff);
                if ((nPecBeg >= 0) && (nPecDiff <= nSmoothWidth / 2))
                    break;
                nRowBeg += 1;
                nPecBeg = (int)floor(m_lstPePure.at(nRowBeg).m_dfWorm);
                }
            // dito for last data sample. nRowEnd is exclusive.
            // watch of end of data, and make sure that nRowEnd is above nRowMid
            while (nRowEnd < m_lstPePure.size())
                {
                int nPecDiff = abs(nPecMid - nPecEnd);
                if (nPecDiff > PecData::c_nRowNum / 2)
                    nPecDiff = abs(PecData::c_nRowNum - nPecDiff);
                if ((nPecEnd >= 0) && (nRowEnd > nRowMid) && (nPecDiff >= (nSmoothWidth - 1) / 2))
                    break;
                nPecEnd = (int)floor(m_lstPePure.at(nRowEnd).m_dfWorm);
                nRowEnd += 1;
                }
            LOG(5, "    [%d,%d,%d[: %d,%d,%d", nRowBeg, nRowMid, nRowEnd, nPecBeg, nPecMid, nPecEnd);

            PePureSample oSample;
            oSample.m_dfWorm = (double)nPecMid;
            if (nRowEnd - nRowBeg <= 1)
                {
                // one single sample: use directly
                oSample.m_dfPe = m_lstPePure.at(nRowMid).m_dfPe;
                }
            else if (bAverage)
                {
                // Average
                double dfPe = 0.;
                for (int nRow = nRowBeg; nRow < nRowEnd; nRow++)
                    dfPe += m_lstPePure.at(nRow).m_dfPe;
                oSample.m_dfPe = dfPe / (double)(nRowEnd - nRowBeg);
                }
            else
                {
                // Median
                double dfPe = 0.;
                QList<double> oList;
                for (int nRow = nRowBeg; nRow < nRowEnd; nRow++)
                    oList.append(m_lstPePure.at(nRow).m_dfPe);
                qSort(oList);
                dfPe = oList.at((nRowEnd - nRowBeg) / 2);
                oSample.m_dfPe = dfPe;
                }
            m_lstPeSmoothed.append(oSample);
            }
        }

    // then combine all raw samples for one PEC-point and smooth again
    // don't work on smoothed data to avoid problems with gaps and such ...
    QVector<QList<double> > aPeSorted;
    aPeSorted.resize(PecData::c_nRowNum);
    for (int nRow = 0; nRow < m_lstPePure.size(); nRow++)
        {
        double dfWormPos = m_lstPePure.at(nRow).m_dfWorm;
        double dfPe = m_lstPePure.at(nRow).m_dfPe;
        int nWormPos = (int)floor(dfWormPos);
        if ((nWormPos >= 0) && (nWormPos < PecData::c_nRowNum))
            aPeSorted[nWormPos].append(dfPe);
        }
    for (int nRow = 0; nRow < PecData::c_nRowNum; nRow++)
        {
        QList<double> lstPe;
        int nSmoothMin = - nSmoothWidth / 2;
        int nSmoothMax = (nSmoothWidth + 1) / 2;
        if (nSmoothMax < 1)
            nSmoothMax = 1;
        for (int nSmooth = nSmoothMin ; nSmooth < nSmoothMax; nSmooth++)
            lstPe += aPeSorted[(nRow + nSmooth + PecData::c_nRowNum) % PecData::c_nRowNum];
        double dfPe = 0.;
        if (lstPe.isEmpty())
            {
            // currently we rely on the user choosing adequate smoothing
            // todo: interpolate?
            }
        else if (bAverage)
            {
            // Average
            for (int nCnt = 0; nCnt < lstPe.size(); nCnt++)
                dfPe += lstPe.at(nCnt);
            dfPe /= (double)(lstPe.size());
            }
        else
            {
            // Median
            qSort(lstPe);
            dfPe = lstPe.at((lstPe.size()) / 2);
            }
        m_aPeProcessed[nRow] = dfPe;
        }

    // if in Touchup-mode, add current PEC table
    bool bTouchup = (m_pUi->peProcMode->currentIndex() == 1);  // 0=Normal, 1=Touchup
    if (bTouchup)
        {
        for (int nRow = 0; nRow < PecData::c_nRowNum; nRow++)
            m_aPeProcessed[nRow] += (double)m_oPecTable[nRow].m_nPos * m_dfPecPosScale;
        }

    m_pPeProcessedModel->setColumnCount(2);
    m_pPeProcessedModel->setRowCount(PecData::c_nRowNum);

    QStringList qslHeaders;
    qslHeaders << tr("Worm Position");
    qslHeaders << tr("PE");
    m_pPeProcessedModel->setHorizontalHeaderLabels(qslHeaders);

    for (int nRow = 0; nRow < PecData::c_nRowNum; nRow++)
        {
        QStandardItem *pItem;
        pItem = new QStandardItem(QString::number(nRow));
        pItem->setEditable(false);
        pItem->setSelectable(false);
        m_pPeProcessedModel->setItem(nRow, 0, pItem);
        pItem = new QStandardItem(QString::number(m_aPeProcessed.at(nRow)));
        pItem->setEditable(false);
        pItem->setSelectable(false);
        m_pPeProcessedModel->setItem(nRow, 1, pItem);
        }

    // m_lstPeriod is the list of periods.
    // each entry holds first and last sample of the period
    // (as index into the list of smoothed PE data)
    // and minimum and maximum PE value of period
    m_lstPeriod.clear();

    Period oPeriod;
    oPeriod.m_nRowBeg = -1;
    double dfWormOld = (double)PecData::c_nRowNum * 4.;
    for (int nRow = 0; nRow < m_lstPeSmoothed.size(); nRow++)
        {
        oPeriod.m_nRowEnd = nRow;
        double dfWormPos = m_lstPeSmoothed.at(nRow).m_dfWorm;
        // a leap of WormPos indicates end of current period
        if (fabs(dfWormPos - dfWormOld) > (double)PecData::c_nRowNum / 2.)
            {
            if (oPeriod.m_nRowBeg >= 0)
                m_lstPeriod.append(oPeriod);
            oPeriod.m_nRowBeg = nRow;
            oPeriod.m_dfPeMin = 1e10;
            oPeriod.m_dfPeMax = -1e10;
            }
        double dfPe = m_lstPeSmoothed.at(nRow).m_dfPe;
        if (oPeriod.m_dfPeMin > dfPe)
            oPeriod.m_dfPeMin = dfPe;
        if (oPeriod.m_dfPeMax < dfPe)
            oPeriod.m_dfPeMax = dfPe;
        dfWormOld = dfWormPos;
        }
    if (oPeriod.m_nRowBeg >= 0)
        {
        oPeriod.m_nRowEnd += 1;
        m_lstPeriod.append(oPeriod);
        }

    recalcPeChart();
    }




/********************************************************************/
/*
    PEC profile
*/


void MainWindow::on_pecProfileOffset_valueChanged(double dfOffset)
    {
    LOG(4, "MainWindow::on_pecProfileOffset_valueChanged");
    Q_UNUSED(dfOffset)
    recalcPeChart();
    }


void MainWindow::on_pecProfileDrift_valueChanged(double dfDrift)
    {
    LOG(4, "MainWindow::on_pecProfileDrift_valueChanged");
    Q_UNUSED(dfDrift)
    recalcPeChart();
    }


void MainWindow::on_pecProfileOpen_clicked()
    {
    LOG(4, "MainWindow::on_pecProfileLoad_clicked");
    QFileDialog oFileDialog(this);
    oFileDialog.setAcceptMode(QFileDialog::AcceptOpen);
    oFileDialog.setFileMode(QFileDialog::ExistingFile);
    oFileDialog.setViewMode(QFileDialog::Detail);
    oFileDialog.setWindowTitle(tr("Open PEC Profile File"));
    oFileDialog.setDirectory(m_oCurrentDirectory);
    QStringList qslFilters;
    qslFilters << tr("PEC Profile File (*.txt)");
    qslFilters << tr("All Files (*)");
    oFileDialog.setFilters(qslFilters);

    QStringList qslFileNames;
    if (oFileDialog.exec())
        {
        m_oCurrentDirectory = oFileDialog.directory();
        qslFileNames = oFileDialog.selectedFiles();
        }
    if (qslFileNames.size() > 0)
        {
        QString qsFileName = qslFileNames.at(0);
        LOG(3, "%s", qsFileName.toLocal8Bit().constData());
        openPecProfileFile(qsFileName);
        }
    }


void MainWindow::openPecProfileFile(const QString &qsFileName)
    {
    LOG(4, "MainWindow::openPecProfileFile");

    QFile oFile(qsFileName);
    oFile.open(QIODevice::ReadOnly | QIODevice::Text);
    QTextStream oStream(&oFile);
    QString qsHeader = oStream.readLine();
    if (qsHeader != QString::fromUtf8("Time, Worm, Wheel, PEC"))
        {
        LOG(0, "MainWindow::openPecProfileFile(): wrong file format");
        return;
        }
    m_lstPecProfile.clear();
    while (!oStream.atEnd())
        {
        QString qsLine = oStream.readLine();
        LOG(5, "    %s", qsLine.toLocal8Bit().constData());
        QStringList qslEntryStrings = qsLine.split(",");
        PeRawSample oSample;
        if ((qslEntryStrings.size() < 4))
            continue;
        oSample.m_oTime = QTime::fromString(qslEntryStrings.at(0), QString::fromUtf8("hh:mm:ss.zzz"));
        oSample.m_dfWorm = qslEntryStrings.at(1).toDouble();
        oSample.m_dfDec = qslEntryStrings.at(2).toDouble();
        oSample.m_dfRa = qslEntryStrings.at(3).toDouble();
        m_lstPecProfile.append(oSample);
        }
    oFile.close();

    m_pPecProfileModel->clear();
    m_pPecProfileModel->setColumnCount(4);
    m_pPecProfileModel->setRowCount(m_lstPecProfile.size());

    QStringList qslHeaders;
    qslHeaders << tr("Time Stamp");
    qslHeaders << tr("Worm Pos");
    qslHeaders << tr("Wheel Pos");
    qslHeaders << tr("PEC Profile");
    m_pPecProfileModel->setHorizontalHeaderLabels(qslHeaders);

    for (int nRow = 0; nRow < m_lstPecProfile.size(); nRow++)
        {
        PeRawSample oSample = m_lstPecProfile.at(nRow);
        QStandardItem *pItem;
        pItem = new QStandardItem(oSample.m_oTime.toString(QString::fromUtf8("hh:mm:ss.zzz")));
        pItem->setEditable(false);
        pItem->setSelectable(false);
        m_pPecProfileModel->setItem(nRow, 0, pItem);
        pItem = new QStandardItem(QString::number(oSample.m_dfWorm, 'f', 3));
        pItem->setEditable(false);
        pItem->setSelectable(false);
        m_pPecProfileModel->setItem(nRow, 1, pItem);
        pItem = new QStandardItem(QString::number(oSample.m_dfDec, 'f', 3));
        pItem->setEditable(false);
        pItem->setSelectable(false);
        m_pPecProfileModel->setItem(nRow, 2, pItem);
        pItem = new QStandardItem(QString::number(oSample.m_dfRa, 'f', 3));
        pItem->setEditable(false);
        pItem->setSelectable(false);
        m_pPecProfileModel->setItem(nRow, 3, pItem);
        }

    recalcPeChart();
    }


void MainWindow::on_pecProfileSave_clicked()
    {
    LOG(4, "MainWindow::on_pecProfileSave_clicked");
    QFileDialog oFileDialog(this);
    oFileDialog.setAcceptMode(QFileDialog::AcceptSave);
    oFileDialog.setFileMode(QFileDialog::AnyFile);
    oFileDialog.setViewMode(QFileDialog::Detail);
    oFileDialog.setWindowTitle(tr("Save PEC Profile As File"));
    oFileDialog.setDirectory(m_oCurrentDirectory);
    QStringList qslFilters;
    qslFilters << tr("PEC Profile File (*.txt)");
    qslFilters << tr("All Files (*)");
    oFileDialog.setFilters(qslFilters);
    oFileDialog.setDefaultSuffix(QString::fromUtf8("txt"));

    QStringList qslFileNames;
    if (oFileDialog.exec())
        {
        m_oCurrentDirectory = oFileDialog.directory();
        qslFileNames = oFileDialog.selectedFiles();
        }
    if (qslFileNames.size() > 0)
        {
        QString qsFileName = qslFileNames.at(0);
        LOG(3, "%s", qsFileName.toLocal8Bit().constData());
        savePecProfileFile(qsFileName);
        }
    }


void MainWindow::savePecProfileFile(const QString &qsFileName)
    {
    LOG(4, "MainWindow::savePecProfileFile");

    QFile oFile(qsFileName);
    // don't use flag QIODevice::Text here, because we need to write CR/LF on all platforms
    oFile.open(QIODevice::WriteOnly);
    QTextStream oStream(&oFile);
    oStream << QString::fromUtf8("Time, Worm, Wheel, PEC\r\n");

    for (int nRow = 0; nRow < m_lstPecProfile.size(); nRow++)
        {
        PeRawSample oSample = m_lstPecProfile.at(nRow);
        oStream << QString::fromUtf8("%1, %2, %3, %4\r\n")
            .arg(oSample.m_oTime.toString(QString::fromUtf8("hh:mm:ss.zzz")))
            .arg(oSample.m_dfWorm, 0, 'f', 3)
            .arg(oSample.m_dfDec, 0, 'f', 3)
            .arg(oSample.m_dfRa, 0, 'f', 3);
        }
    oStream << QString::fromUtf8("\r\n");
    oFile.close();
    }


void MainWindow::on_mcuGuideEast_toggled(bool bOn)
    {
    if (bOn && (m_eGuideDir != east))
        {
        // start moving east at guide rate
        m_eGuideDir = east;
        m_pUi->mcuGuideWest->setChecked(false);
        McuRawCommand oCmd1;
        oCmd1.setCommand(QString::fromUtf8("#:Q#").toLatin1());
        McuRawCommand oCmd2;
        oCmd2.setCommand(QString::fromUtf8("#:Me#").toLatin1());
        if (mcuCommunicationBegin())
            {
            oCmd1.mcuSend(*m_pMcuPort);
            oCmd2.mcuSend(*m_pMcuPort);
            mcuCommunicationEnd();
            }
        }
    else if ((!bOn) && (m_eGuideDir == east))
        {
        // stop moving
        m_eGuideDir = stop;
        McuRawCommand oCmd1;
        oCmd1.setCommand(QString::fromUtf8("#:Q#").toLatin1());
        if (mcuCommunicationBegin())
            {
            oCmd1.mcuSend(*m_pMcuPort);
            mcuCommunicationEnd();
            }
        }
    }


void MainWindow::on_mcuGuideWest_toggled(bool bOn)
    {
    if (bOn && (m_eGuideDir != west))
        {
        // start moving east at guide rate
        m_eGuideDir = west;
        m_pUi->mcuGuideEast->setChecked(false);
        McuRawCommand oCmd1;
        oCmd1.setCommand(QString::fromUtf8("#:Q#").toLatin1());
        McuRawCommand oCmd2;
        oCmd2.setCommand(QString::fromUtf8("#:Mw#").toLatin1());
        if (mcuCommunicationBegin())
            {
            oCmd1.mcuSend(*m_pMcuPort);
            oCmd2.mcuSend(*m_pMcuPort);
            mcuCommunicationEnd();
            }
        }
    else if ((!bOn) && (m_eGuideDir == west))
        {
        // stop moving
        m_eGuideDir = stop;
        McuRawCommand oCmd1;
        oCmd1.setCommand(QString::fromUtf8("#:Q#").toLatin1());
        if (mcuCommunicationBegin())
            {
            oCmd1.mcuSend(*m_pMcuPort);
            mcuCommunicationEnd();
            }
        }
    }


/*!
    \todo show message box if opening port failed
*/

void MainWindow::on_mcuStart_toggled(bool bOn)
    {
    LOG(4, "MainWindow::on_mcuStart_toggled");

    if (bOn)
        {
        clearPecProfile();
        startPecProfile(true);
        m_pUi->mcuStart->setText(QString::fromUtf8("Stop"));
        m_pUi->mcuPause->setText(QString::fromUtf8("Pause"));
        m_pUi->mcuPause->setChecked(false);
        m_pUi->mcuPause->setEnabled(true);
        }
    else
        {
        stopPecProfile();
        m_pUi->mcuPause->setEnabled(false);
        m_pUi->mcuPause->setChecked(false);
        m_pUi->mcuPause->setText(QString::fromUtf8("Pause"));
        m_pUi->mcuStart->setText(QString::fromUtf8("Start"));
        }
    }


void MainWindow::on_mcuPause_toggled(bool bOn)
    {
    LOG(4, "MainWindow::on_mcuPause_toggled");

    if (!m_pUi->mcuStart->isChecked())
        return;

    if (bOn)
        {
        stopPecProfile();
        m_pUi->mcuPause->setText(QString::fromUtf8("Continue"));
        }
    else
        {
        startPecProfile(false);
        m_pUi->mcuPause->setText(QString::fromUtf8("Pause"));
        }
    }


void MainWindow::clearPecProfile()
    {
    // clear PEC profile listview and graph
    m_lstPecProfile.clear();

    m_pPecProfileModel->clear();
    m_pPecProfileModel->setColumnCount(4);

    QStringList qslHeaders;
    qslHeaders << tr("Time Stamp");
    qslHeaders << tr("Worm Pos");
    qslHeaders << tr("Wheel Pos");
    qslHeaders << tr("PEC Profile");
    m_pPecProfileModel->setHorizontalHeaderLabels(qslHeaders);

    recalcPeChart();
    }


void MainWindow::startPecProfile(bool bSetSpeed)
    {
    if (mcuCommunicationBegin())
        {
        if (bSetSpeed)
            {
            // switch to sidereal tracking rate, and switch intellytrack off
            McuDataTracking oTracking;
            oTracking.setRate(McuDataTracking::sidereal);
            McuDataPecState oPec;
            oPec.setIntellyTrack(false);
            oTracking.mcuSet(*m_pMcuPort);
            oPec.mcuSetIntelly(*m_pMcuPort);
            }
        m_pMcuTimer->start((int)(m_pUi->mcuPoll->value() * 1000.));
        }
    }


void MainWindow::stopPecProfile()
    {
    m_pMcuTimer->stop();
    mcuCommunicationEnd();
    }


void MainWindow::mcuTimer()
    {
    LOG(4, "MainWindow::mcuTimer");

    McuDataPecState oPecState;
    oPecState.mcuGet(*m_pMcuPort);
    m_pUi->mcuStatusPec->setText(oPecState.state());

    PeRawSample oSample;
    oSample.m_oTime = oPecState.time();
    oSample.m_dfWorm = oPecState.wormPos();
    if (m_lstPecProfile.size() == 0)
        {
        oSample.m_dfRa = 0.;
        oSample.m_dfDec = 0.;
        }
    else
        {
        double dfWormWheelTeeth = (double)m_nWormWheelTeeth;
        double dfTime = (double)(m_lstPecProfile.at(0).m_oTime.msecsTo(oSample.m_oTime)) / 1000.;
        double dfWorm = (oSample.m_dfWorm - m_lstPecProfile.at(0).m_dfWorm) * (-1.);
        dfWorm -= floor(dfWorm / (double)PecData::c_nRowNum) * (double)PecData::c_nRowNum;
        double dfWheel = dfTime / (m_dfSiderealDay / dfWormWheelTeeth / (double)PecData::c_nRowNum);
        dfWheel -= floor(dfWheel / (double)PecData::c_nRowNum) * (double)PecData::c_nRowNum;
        double dfWormError = dfWorm - dfWheel;
        dfWormError -= floor(dfWormError / (double)PecData::c_nRowNum + .5) * (double)PecData::c_nRowNum;
        oSample.m_dfRa = (dfWormError) * (360. * 3600. / dfWormWheelTeeth / (double)PecData::c_nRowNum);
        oSample.m_dfDec = dfWheel; // abuse Dec as storage for wheel position
        }
    m_lstPecProfile.append(oSample);

    QList<QStandardItem *> oItemList;
    oItemList.append(new QStandardItem(oSample.m_oTime.toString(QString::fromUtf8("hh:mm:ss.zzz"))));
    oItemList.append(new QStandardItem(QString::number(oSample.m_dfWorm, 'f', 3)));
    oItemList.append(new QStandardItem(QString::number(oSample.m_dfDec, 'f', 3)));
    oItemList.append(new QStandardItem(QString::number(oSample.m_dfRa, 'f', 3)));
    m_pPecProfileModel->appendRow(oItemList);
    m_pUi->viewPecProfile->scrollTo(oItemList.at(0)->index());
    LOG(3, "%s %.3f %.3f %.3f",
        LQS(oSample.m_oTime.toString(QString::fromUtf8("hh:mm:ss.zzz"))),
        oSample.m_dfWorm, oSample.m_dfDec, oSample.m_dfRa);

    recalcPeChart();
    }




/********************************************************************/
/*
    Charts
*/


/*!
    Recalculate and redraw all kind of charts

    \todo move this out of MainWindow
    \todo implement PEC profile graph properly
*/

void MainWindow::recalcPeChart()
    {
    LOG(4, "MainWindow::recalcPeChart");

    m_pUi->chartPe->clear();

    bool bChartPeRaw = m_pUi->toggleChartPeRaw->isChecked();
    bool bChartPeProcessed = m_pUi->toggleChartPeProcessed->isChecked();
    bool bChartPec = m_pUi->toggleChartPec->isChecked();
    bool bChartPecProfile = m_pUi->toggleChartPecProfile->isChecked();
    bool bChartPeResidual = m_pUi->toggleChartPeResidual->isChecked();

    double dfPecOffset = -m_pUi->pecGraphOffset->value() / m_dfPecPosScale;
#if 0
    bool bChartPe = bChartPeRaw || bChartPeProcessed || bChartPeResidual;
    int nPeMin = 1000;
    int nPeMax = -1000;
    for (int nPeriod = 0; nPeriod < m_lstPeriod.size(); nPeriod++)
        {
        if (nPeMin > (int)floor(m_lstPeriod.at(nPeriod).m_dfPeMin))
            nPeMin = (int)floor(m_lstPeriod.at(nPeriod).m_dfPeMin);
        if (nPeMax < (int)ceil(m_lstPeriod.at(nPeriod).m_dfPeMax))
            nPeMax = (int)ceil(m_lstPeriod.at(nPeriod).m_dfPeMax);
        }

    int nPecMin = (int)floor(dfPecOffset);
    int nPecMax = (int)ceil(dfPecOffset);
    for (int nRow = 0; nRow < PecData::c_nRowNum; nRow++)
        {
        double dfPos = ((double)m_oPecTable[nRow].m_nPos + dfPecOffset) * m_dfPecPosScale;
        if (nPecMin > (int)floor(dfPos))
            nPecMin = (int)floor(dfPos);
        if (nPecMax < (int)ceil(dfPos))
            nPecMax = (int)ceil(dfPos);
        }

    int nXMin = 0;
    int nXMax = PecData::c_nRowNum;
    int nYMin = 0;
    int nYMax = 0;
    if (bChartPe)
        {
        if (nYMin > nPeMin)
            nYMin = nPeMin;
        if (nYMax < nPeMax)
            nYMax = nPeMax;
        }
    if (bChartPec)
        {
        if (nYMin > nPecMin)
            nYMin = nPecMin;
        if (nYMax < nPecMax)
            nYMax = nPecMax;
        }
    int nYRange = nYMax - nYMin;
    if (nYRange < 10)
        {
        nYMin -= (10 - nYRange) / 2;
        nYMax += (10 - nYRange + 1) / 2;
        }
    if (nYRange > 300)
        {
        nYMin += (nYRange - 300) / 2;
        nYMax -= (nYRange - 300 + 1) / 2;
        }

    double dfXScale = 500. / (double)(nXMax - nXMin + 1);
    double dfYScale = 500. / (double)(nYMax - nYMin + 1);
    double dfYOffsSpd = dfYZeroSpd - dfYScale * (double)nYMin / dfYScaleSpd;
    double dfYOffsAcc = dfYZeroAcc - dfYScale * (double)nYMin / dfYScaleAcc;
#endif
    Graph oGraphPeRaw(Graph::line);
    Graph oGraphPeProcessed(Graph::line);
    Graph oGraphPos(Graph::line);
    Graph oGraphSpd((m_oPecTable.traversal() == PecData::forward)?
        Graph::stairLeft: Graph::stairRight);
    Graph oGraphAcc(Graph::tick);
    Graph oGraphPeProfile(Graph::line);

    Grid oGridPe;
    Grid oGridPecPos;
    Grid oGridPecSpd;
    Grid oGridPecAcc;

    if (bChartPeRaw)
        {
        for (int nPeriod = 0; nPeriod < m_lstPeriod.size(); nPeriod++)
            {
            int nRowBeg = m_lstPeriod.at(nPeriod).m_nRowBeg;
            int nRowEnd = m_lstPeriod.at(nPeriod).m_nRowEnd;
            LOG(4, "MainWindow::recalcPeChart: period %d: [%d..%d]", nPeriod, nRowBeg, nRowEnd);
            for (int nRow = nRowBeg; nRow < nRowEnd; nRow++)
                oGraphPeRaw.addData(
                    m_lstPeSmoothed.at(nRow).m_dfWorm,
                    m_lstPeSmoothed.at(nRow).m_dfPe,
                    (nRow == nRowBeg));
            }
        oGridPe.addData((double)PecData::c_nRowNum, 0.);
        oGridPe.addGraph(oGraphPeRaw);
        }

    if (bChartPeProcessed)
        {
        for (int nRow = 0; nRow <= PecData::c_nRowNum; nRow++)
            oGraphPeProcessed.addData((double)nRow, m_aPeProcessed.at(nRow % PecData::c_nRowNum));
        oGridPe.addData((double)PecData::c_nRowNum, 0.);
        oGridPe.addGraph(oGraphPeProcessed);
        }

    if (bChartPec)
        {
        for (int nRow = 0; nRow < PecData::c_nRowNum; nRow++)
            {
            oGraphPos.addData((double)nRow,
                (double)m_oPecTable[nRow].m_nPos * m_dfPecPosScale - dfPecOffset);
            oGraphSpd.addData((double)nRow, (double)m_oPecTable[nRow].m_nSpd);
            switch (m_oPecTable[nRow].m_eVal)
                {
                case PecSample::keep:
                    break;
                case PecSample::faster:
                    oGraphAcc.addData((double)nRow + .5, .75);
                    break;
                case PecSample::slower:
                    oGraphAcc.addData((double)nRow + .5, -.75);
                    break;
                case PecSample::reset:
                    oGraphAcc.addData((double)nRow + .5, 0.);
                    break;
                }
            }
        oGraphPos.addData((double)PecData::c_nRowNum,
            (double)m_oPecTable[0].m_nPos * m_dfPecPosScale - dfPecOffset);
        oGraphSpd.addData((double)PecData::c_nRowNum, (double)m_oPecTable[0].m_nSpd);
        oGridPe.addGraph(oGraphPos);
        oGridPecPos.addData((double)PecData::c_nRowNum, 0.);
        oGridPecPos.addGraph(oGraphPos);
        }
    // always reserve space for PEC acc/spd
    oGridPecSpd.addData(0., -(double)PecData::c_nFactorMax);
    oGridPecSpd.addData((double)PecData::c_nRowNum, (double)PecData::c_nFactorMax);
    oGridPecSpd.addGraph(oGraphSpd);
    oGridPecAcc.addData(0., -1.);
    oGridPecAcc.addData((double)PecData::c_nRowNum, 1.);
    oGridPecAcc.addGraph(oGraphAcc);

    if ((bChartPecProfile) && (m_lstPecProfile.size() > 0))
        {
        int nRow = (int)floor(m_lstPecProfile.at(0).m_dfWorm);
        double dfPecProfileOffset = ((double)m_oPecTable[nRow].m_nPos + dfPecOffset) * m_dfPecPosScale;
        double dfXStart = (m_lstPecProfile.size() > 0)? m_lstPecProfile.at(0).m_dfWorm: 0.;
        double dfXPrev = dfXStart;
        double dfXPeriod = 0.;
        for (int nRow = 0; nRow < m_lstPecProfile.size(); nRow++)
            {
            double dfX = m_lstPecProfile.at(nRow).m_dfWorm;
            bool bWrap = (fabs(dfX - dfXPrev) > (double)PecData::c_nRowNum / 2.);
            if (bWrap)
                dfXPeriod += (double)PecData::c_nRowNum * ((dfX > dfXPrev)? -1.: 1.);
            double dfY = dfPecProfileOffset + m_pUi->pecProfileOffset->value() -
                m_lstPecProfile.at(nRow).m_dfRa -
                (dfX + dfXPeriod - dfXStart) * m_pUi->pecProfileDrift->value() / (double)PecData::c_nRowNum;
            oGraphPeProfile.addData(dfX, dfY, bWrap);
            dfXPrev = dfX;
            }
        oGridPe.addData((double)PecData::c_nRowNum, 0.);
        oGridPe.addGraph(oGraphPeProfile);
        }

    if (bChartPeResidual)
        {
        }

    // limit y range of main chart to give good appearance and
    // avoid insane y values hanging the QGraphicsView
    double dfRangeX = oGridPe.maxX() - oGridPe.minX();
    double dfRangeY = oGridPe.maxYSanitized() - oGridPe.minYSanitized();
    // scale main chart to same x/y resolution, so that numbers look good
    // move main chart down such that top edge is at scene coordinate 0
    static const double c_dfScale = 450.;
    static const double c_dfBorder = 25.;
    double dfXScale = c_dfScale / dfRangeX;
    double dfYScalePos = c_dfScale / dfRangeY;
    double dfYOffsPos = oGridPe.maxYSanitized();
    // scale speed chart to same x/y resolution
    // move speed chart below main chart, considering that maximum value is PecData::c_nFactorMax
    double dfYScaleSpd = c_dfScale / dfRangeX;
    double dfYOffsSpd = (c_dfScale + c_dfBorder) / dfYScaleSpd + (double)PecData::c_nFactorMax;
    // scale acceleration chart to four times x/y resolution (for readablility)
    // move acceleration chart below speed chart, considering that maximum value is 1
    double dfYScaleAcc = c_dfScale / dfRangeX * 4.;
    double dfYOffsAcc = (c_dfScale + c_dfBorder + dfYScaleSpd * 2. * (double)PecData::c_nFactorMax +
        c_dfBorder) / dfYScaleAcc + 1.;

    oGraphPeRaw.render(m_pPeChart,
        dfXScale, dfYOffsPos, dfYScalePos,
        2.);
    oGraphPeProcessed.render(m_pPeChart,
        dfXScale, dfYOffsPos, dfYScalePos,
        3., QPen(Qt::red));
    oGraphPos.render(m_pPeChart,
        dfXScale, dfYOffsPos, dfYScalePos,
        4., QPen(Qt::darkGreen));
    oGraphSpd.render(m_pPeChart,
        dfXScale, dfYOffsSpd, dfYScaleSpd,
        4., QPen(Qt::darkGreen));
    oGraphAcc.render(m_pPeChart,
        dfXScale, dfYOffsAcc, dfYScaleAcc,
        4., QPen(Qt::darkGreen));
    oGraphPeProfile.render(m_pPeChart,
        dfXScale, dfYOffsPos, dfYScalePos,
        3.5, QPen(QColor(170, 0, 255)));

    oGridPe.render(m_pPeChart,
        dfXScale, dfYOffsPos, dfYScalePos,
        Grid::both, Grid::both, Grid::min, Grid::min,
        32., 8., 5., 1.);
    oGridPecPos.render(m_pPeChart,
        dfXScale, dfYOffsPos, dfYScalePos,
        Grid::none, Grid::none, Grid::undec, Grid::max,
        32., 8., 5., 1., -1., -2., Qt::darkGreen);
    oGridPecSpd.render(m_pPeChart,
        dfXScale, dfYOffsSpd, dfYScaleSpd,
        Grid::both, Grid::minor, Grid::min, Grid::max,
        32., 8., (double)PecData::c_nFactorMax, (double)PecData::c_nFactorMax,
        -1., -2., Qt::darkGreen);
    oGridPecAcc.render(m_pPeChart,
        dfXScale, dfYOffsAcc, dfYScaleAcc,
        Grid::both, Grid::minor, Grid::undec, Grid::max,
        32., 8., 5., 1., -1., -2., Qt::darkGreen);

    m_pUi->chartPe->requestSceneRender();
    }

